use ldap3::{LdapConn, Scope, SearchEntry};
use skynet_ldap_backend::{db_init, get_config, AccountWolves, Accounts, Config};
use sqlx::{Pool, Sqlite};
use wolves_oxidised::WolvesUser;

#[async_std::main]
async fn main() -> tide::Result<()> {
  let config = get_config();
  let db = db_init(&config).await.unwrap();

  update_wolves(&config, &db).await;
  update_ldap(&config, &db).await;

  Ok(())
}

async fn update_wolves(config: &Config, db: &Pool<Sqlite>) {
  let wolves = wolves_oxidised::Client::new(&config.wolves_url, None);
  for account in wolves.get_members(&config.wolves_key).await {
    update_account(db, &account).await;
  }
}

async fn update_ldap(config: &Config, db: &Pool<Sqlite>) {
  let mut ldap = match LdapConn::new(&config.ldap_host) {
    Ok(s) => s,
    Err(e) => {
      println!("{:?}", e);
      return;
    }
  };

  match ldap.simple_bind(&config.ldap_admin, &config.ldap_admin_pw) {
    Ok(_) => {}
    Err(e) => {
      println!("{:?}", e);
      return;
    }
  }

  // use this to pre load a large chunk of data
  if let Ok(x) = ldap.search(
    "ou=users,dc=skynet,dc=ie",
    Scope::OneLevel,
    "(objectClass=*)",
    vec!["uid", "uidNumber", "skDiscord", "skMemberOf", "mail", "skID", "userPassword"],
  ) {
    if let Ok((rs, _res)) = x.success() {
      for entry in rs {
        let tmp = SearchEntry::construct(entry);

        let mut tmp_account = Accounts {
          user: "".to_string(),
          uid: 0,
          mail: "".to_string(),
          student_id: "".to_string(),
          secure: false,
        };

        // pull out the required info
        if tmp.attrs.contains_key("uid") && !tmp.attrs["uid"].is_empty() {
          tmp_account.user = tmp.attrs["uid"][0].clone();
        }
        if tmp.attrs.contains_key("uidNumber") && !tmp.attrs["uidNumber"].is_empty() {
          tmp_account.uid = tmp.attrs["uidNumber"][0].clone().parse().unwrap_or(0);
        }
        if tmp.attrs.contains_key("mail") && !tmp.attrs["mail"].is_empty() {
          tmp_account.mail = tmp.attrs["mail"][0].clone();
        }
        if tmp.attrs.contains_key("skID") && !tmp.attrs["skID"].is_empty() {
          tmp_account.student_id = tmp.attrs["skID"][0].clone();
        }
        if tmp.attrs.contains_key("userPassword") && !tmp.attrs["userPassword"].is_empty() {
          tmp_account.secure = tmp.attrs["userPassword"][0].starts_with("{SSHA512}")
        }

        if !tmp_account.user.is_empty() {
          sqlx::query_as::<_, Accounts>(
            "
            INSERT OR REPLACE INTO accounts (user, uid, mail, student_id, secure)
            VALUES (?1, ?2, ?3, ?4, ?5)
            ",
          )
          .bind(&tmp_account.user)
          .bind(tmp_account.uid)
          .bind(&tmp_account.mail)
          .bind(&tmp_account.student_id)
          .bind(tmp_account.secure)
          .fetch_optional(db)
          .await
          .ok();
        }
      }
    }
  }

  // done with ldap
  ldap.unbind().unwrap();
}

async fn update_account(db: &Pool<Sqlite>, account: &WolvesUser) {
  sqlx::query_as::<_, AccountWolves>(
    "
    INSERT OR REPLACE INTO accounts_wolves (id_wolves, id_student, email, expiry, name_first, name_second)
    VALUES (?1, ?2, ?3, ?4, ?5, ?6)
    ",
  )
  .bind(&account.member_id)
  .bind(&account.student_id)
  .bind(&account.contact_email)
  .bind(&account.expiry)
  .bind(&account.first_name)
  .bind(&account.last_name)
  .fetch_optional(db)
  .await
  .ok();
}
